import os
import glob
import pathlib
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata
from vyperdatum.pipeline import Pipeline


def transform(input_file, home_dir):
    options = {"options": ["s_coord_epoch=2010.0", "t_coord_epoch=2010.0"]}
    warp_kwargs_vertical = {
                            "outputType": gdal.gdalconst.GDT_Float32,
                            "srcBands": [1],
                            "dstBands": [1],
                            "warpOptions": ["APPLY_VERTICAL_SHIFT=YES", "SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"],
                            "errorThreshold": 0,
                            }
    
    t1 = Transformer(crs_from="EPSG:32609",
                     crs_to="EPSG:9755",
                     allow_ballpark=False
                     )
    out_file1 = pathlib.Path(input_file).with_stem("_01_" + pathlib.Path(input_file).stem)
    t1.transform_raster(input_file=input_file,
                        output_file=out_file1,
                        apply_vertical=False,
                        warp_kwargs=options
                        )

    t2 = Transformer(crs_from="EPSG:9755",
                     crs_to="EPSG:6318",
                     allow_ballpark=False
                     )
    out_file2 = pathlib.Path(input_file).with_stem("_02_" + pathlib.Path(input_file).stem)
    t2.transform_raster(input_file=out_file1,
                        output_file=out_file2,
                        apply_vertical=False,
                        warp_kwargs=options
                        )

    t3 = Transformer(crs_from="EPSG:6318",
                     crs_to="EPSG:26909",
                     allow_ballpark=False
                     )
    # out_file3 = pathlib.Path(input_file).with_stem("_03_" + pathlib.Path(input_file).stem)
    p = pathlib.Path(input_file)
    xform_dir = os.path.join(home_dir + r"\Manual", p.parent.name)
    os.makedirs(xform_dir, exist_ok=True)
    out_file3 = os.path.join(xform_dir, p.name)
    t3.transform_raster(input_file=out_file2,
                        output_file=out_file3,
                        apply_vertical=False,
                        )

    os.remove(out_file1)
    os.remove(out_file2)
    return out_file3

def transform_MLLW(input_file, home_dir, in_wgs84_crs, out_nad83_crs):
    options = {"options": ["s_coord_epoch=2010.0", "t_coord_epoch=2010.0"]}
    warp_kwargs_vertical = {
                            "outputType": gdal.gdalconst.GDT_Float32,
                            "srcBands": [1],
                            "dstBands": [1],
                            "warpOptions": ["APPLY_VERTICAL_SHIFT=YES"],
                            "errorThreshold": 0,
                            }

    t1 = Transformer(crs_from=f"{in_wgs84_crs}+EPSG:5866",
                     crs_to="EPSG:9755+EPSG:5866",
                     allow_ballpark=False
                     )
    out_file1 = pathlib.Path(input_file).with_stem("_01_" + pathlib.Path(input_file).stem)
    t1.transform_raster(input_file=input_file,
                        output_file=out_file1,
                        apply_vertical=False,
                        warp_kwargs=options
                        )

    t2 = Transformer(crs_from="EPSG:9755",   # must have been EPSG:9755+EPSG:5866, but bug in db/register
                     crs_to="EPSG:6318",     # must have been EPSG:6318+EPSG:5866, but bug in db/register
                     allow_ballpark=False
                     )
    out_file2 = pathlib.Path(input_file).with_stem("_02_" + pathlib.Path(input_file).stem)
    t2.transform_raster(input_file=out_file1,
                        output_file=out_file2,
                        apply_vertical=False,
                        warp_kwargs=options
                        )

    t3 = Transformer(crs_from="EPSG:6318+EPSG:5866",
                     crs_to=f"{out_nad83_crs}+EPSG:5866",
                     allow_ballpark=False
                     )
    # out_file3 = pathlib.Path(input_file).with_stem("_03_" + pathlib.Path(input_file).stem)
    p = pathlib.Path(input_file)
    xform_dir = os.path.join(home_dir + r"\Manual", p.parent.name)
    os.makedirs(xform_dir, exist_ok=True)
    out_file3 = os.path.join(xform_dir, p.name)
    t3.transform_raster(input_file=out_file2,
                        output_file=out_file3,
                        apply_vertical=False,
                        )

    os.remove(out_file1)
    os.remove(out_file2)
    return out_file3


if __name__ == "__main__":
    # home_dir: path to the dir where the "Original" and "Manual" are stored
    home_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBA"
    files = glob.glob(home_dir + r"\Original\**\W*.bag", recursive=True)
    for i, input_file in enumerate(files):
        print(f"{i+1}/{len(files)}: {input_file}")
        raster_metadata(input_file, verbose=True)
        if os.path.basename(input_file).startswith("H12137"):
            transformed_file = transform_MLLW(input_file, home_dir, "EPSG:32619", "EPSG:26919")
        elif os.path.basename(input_file).startswith("W00656") or os.path.basename(input_file).startswith("D00158"):
            # transformed_file = transform_MLLW(input_file, home_dir, "EPSG:32617", "EPSG:26917")
            crs_from = "EPSG:32617+EPSG:5866"
            crs_to = "EPSG:26917+EPSG:5866"
            print(Pipeline(crs_from=crs_from, crs_to=crs_to).linear_steps())
            # print(Pipeline(crs_from=crs_from, crs_to=crs_to).graph_steps())
            # tf = Transformer(crs_from=crs_from,
            #                  crs_to=crs_to,
            #                  steps=["EPSG:32617+EPSG:5866", "EPSG:9755+EPSG:5866", "EPSG:6318+EPSG:5866", "EPSG:26917+EPSG:5866"]
            #                 #  steps=Pipeline(crs_from=crs_from, crs_to=crs_to).linear_steps()
            #                 #  steps=Pipeline(crs_from=crs_from, crs_to=crs_to).graph_steps()
            #                  )
            # tf.transform_raster(input_file=input_file,
            #                     output_file=input_file.replace("Original", "Manual"),
            #                     overview=False,
            #                     )
        else:
            transformed_file = transform(input_file, home_dir)

        print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



import glob
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
import pyproj as pp




if __name__ == "__main__":

    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBC_mixed\Original\**\*.tiff"
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBC_mixed\Original\**\*.bag"
    # parent_dir = r"C:\Users\mohammad.ashkezari\Desktop\bag\Original\*.bag"
    files = glob.glob(parent_dir, recursive=True)[:]
    files = [r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBC_mixed\Original\E01096\E01096_MB_4m_MLLW_7of7_clipped.bag"]
    files = [r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBC_mixed\Original\E01096\clipped.bag"]
    files = [r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBC_mixed\Original\E01096\E01096_MB_4m_MLLW_7of7.bag"]
    crs_from = "EPSG:32619+NOAA:101"
    crs_to = "EPSG:6348+NOAA:101"
    # steps = [{'crs_from': 'EPSG:32619', 'crs_to': 'EPSG:9755', 'v_shift': False},
    #          {'crs_from': 'EPSG:9755', 'crs_to': 'EPSG:6318', 'v_shift': False},
    #          {'crs_from': 'EPSG:6318', 'crs_to': 'EPSG:6348', 'v_shift': False},
    #          ]
    for i, input_file in enumerate(files):
        print(f"{i+1}/{len(files)}: {input_file}")
        tf = Transformer(crs_from=crs_from,
                         crs_to=crs_to,
                        #  steps=steps
                         )
        output_file = input_file.replace("Original", "Manual")
        tf.transform(input_file=input_file,
                     output_file=output_file,
                     pre_post_checks=True,
                     vdatum_check=False
                     )
        print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



import os
from vyperdatum.transformer import Transformer
from vyperdatum.pipeline import Pipeline




if __name__ == "__main__":
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\RSD\Alaska\Original"
    crs_from = "EPSG:6318+EPSG:7968"
    crs_to = "EPSG:6318+NOAA:98"
    # print(Pipeline(crs_from=crs_from, crs_to=crs_to).graph_steps())

    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                     steps=['EPSG:6318+EPSG:7968', 'EPSG:6319', 'EPSG:7912', 'EPSG:9989', 'EPSG:9990+NOAA:98', 'EPSG:4267']
                     #  steps=Pipeline(crs_from=crs_from, crs_to=crs_to).linear_steps()
                     #  steps=Pipeline(crs_from=crs_from, crs_to=crs_to).graph_steps()
                     )

    # for i, input_file in enumerate(files[:1]):
    #     print(f"{i+1}/{len(files)}: {input_file}")
    #     tf = Transformer(crs_from=crs_from,
    #                      crs_to=crs_to,
    #                      steps=["EPSG:6346", "EPSG:6319", "EPSG:6318+NOAA:5224", "EPSG:6346+NOAA:5224"]
    #                     #  steps=Pipeline(crs_from=crs_from, crs_to=crs_to).linear_steps()
    #                     #  steps=Pipeline(crs_from=crs_from, crs_to=crs_to).graph_steps()
    #                      )
    #     tf.transform_raster(input_file=input_file,
    #                         output_file=input_file.replace("Original", "Manual"),
    #                         overview=False,
    #                         )
    #     print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



import glob
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
import pyproj as pp




if __name__ == "__main__":

    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBC\Original\ME2201_block04*\*.tif"
    files = glob.glob(parent_dir, recursive=True)[:]
    crs_from = "EPSG:6348"
    crs_to = "EPSG:6348+NOAA:101"
    for i, input_file in enumerate(files):
        print(f"{i+1}/{len(files)}: {input_file}")
        tf = Transformer(crs_from=crs_from,
                         crs_to=crs_to
                         )
        output_file = input_file.replace("Original", "Manual")
        tf.transform_raster(input_file=input_file,
                            output_file=output_file,
                            overview=False,
                            pre_post_checks=True,
                            vdatum_check=False
                            )
        print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



import pyproj as pp
from vyperdatum.transformer import Transformer
from pyproj.transformer import TransformerGroup

print(f"proj dat dir: {pp.datadir.get_data_dir()}")

crs_from = f"EPSG:6318+NOAA:25101"
crs_to = f"EPSG:6318+NOAA:98"

t1 = pp.Transformer.from_crs(crs_from=crs_from,
                                crs_to=crs_to,
                                allow_ballpark=True,
                                only_best=True
                                )

# tf = Transformer(crs_from=crs_from,crs_to=crs_to)

tg = TransformerGroup(crs_from=crs_from,
                    crs_to=crs_to,
                    allow_ballpark=True,
                    )

print(f"Number of transformers: {len(tg.transformers)}")


import os
from pathlib import Path
import pyproj as pp
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
from osgeo import gdal
import re
import glob


def get_tiff_files(parent_dir: str, extention: str) -> list:
    tiff_files = []
    for (dirpath, dirnames, filenames) in os.walk(parent_dir):
        for filename in filenames:
            if filename.endswith(extention):
                tiff_files.append(os.sep.join([dirpath, filename]))
    return tiff_files


def transform_with_vyperdatum(input_file, output_file, crs_from, crs_to, steps):
    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                     steps=steps
                     )
    tf.transform_raster(input_file=input_file,
                        output_file=output_file,
                        overview=False,
                        pre_post_checks=True,
                        vdatum_check=False
                        )
    return



def transform_with_concat_pipe(input_file, output_file):
    output_vrt = output_file.replace(".tif", ".vrt")

    ak_vyperdatum_nwld = "nwldatum_4.7.0_20240621"
    ak_vyperdatum_nwld = "nwldatum_4.0.0_20190729" # (NOAA:55) to match up to VDatum results
    ak_zone = 9

    SRC_HORZ_PUSH = "+step +proj=push +v_1 +v_2"
    SRC_HORZ_POP = "+step +proj=pop +v_1 +v_2"

    NAD83_UTM_to_Geographic = lambda zone: f"+step +inv +proj=utm +zone={zone} +ellps=GRS80"
    NAD83_Geographic_to_UTM = lambda zone: f"+step +proj=utm +zone={zone} +ellps=GRS80"

    # projinfo -s EPSG:6318+EPSG:5703 -t EPSG:6319 --spatial-test intersects --hide-ballpark -o PROJ
    # (CONUS is Operation 1, AK is Operation 2)
    AK_NAVD88_GEOID12B_to_NAD83 = "+step +proj=vgridshift +grids=us_noaa_g2012ba0.tif +multiplier=1"
    # CONUS_NAVD88_GEOID18_to_NAD83 = "+step +proj=vgridshift +grids=us_noaa_g2018u0.tif +multiplier=1"


    # Note: NWLD hydroids are <datum> - [NAD83(2011) 2010.0, ITRF2020 2020.0]; to subtract, we inverse vgridshift (+step +inv)
    NAD83_2011_to_NWLD = lambda vdatum, nwld: f"+step +inv +proj=vgridshift +grids=us_noaa_nos_{vdatum}-NAD83(2011)_2010.0_({nwld}).tif +multiplier=1"

    # zone: 9, etc.
    # nwld: "nwldatum_4.7.0_20240621" (AK_ERTDM_2023), "nwldatum_4.0.0_20190729" (VDatum 4.0 SE AK only), "nwldatum_3.7.0_20170907" (AK_ERTDM_2021)
    # vdatum: "underkeel_hydroid", "MLLW", etc.
    AK_NAD83_UTM_NAVD88_height_to_NAD83_UTM_NWLD_height = lambda zone, nwld, vdatum: f"""
        +proj=pipeline
            {NAD83_UTM_to_Geographic(zone)}
            {SRC_HORZ_PUSH}
            {AK_NAVD88_GEOID12B_to_NAD83}
            {NAD83_2011_to_NWLD(vdatum, nwld)}
            {SRC_HORZ_POP}
            {NAD83_Geographic_to_UTM(zone)}
    """
    proj_pipeline_transform = AK_NAD83_UTM_NAVD88_height_to_NAD83_UTM_NWLD_height(ak_zone, ak_vyperdatum_nwld, "MLLW")


    # output pixel resolution = input pixel resolution
    with gdal.Open(input_file, gdal.GA_ReadOnly) as input_ds:
        geotransform = input_ds.GetGeoTransform()
        xres, yres = geotransform[1], geotransform[5]

    # create vrt transformation template dataset
    ds = gdal.Warp(output_vrt, input_file, format="vrt", outputType=gdal.gdalconst.GDT_Float32, warpOptions=["APPLY_VERTICAL_SHIFT=YES", "SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"], errorThreshold=0, xRes=xres, yRes=yres, targetAlignedPixels=True, coordinateOperation=proj_pipeline_transform)

    # remove whitespace formatting used in proj pipeline string and put it in GeoTIFF metadata
    proj_pipeline_transform = re.sub(r'\s{2,}', ' ', proj_pipeline_transform).strip()
    ds.SetMetadataItem('TIFFTAG_IMAGEDESCRIPTION', proj_pipeline_transform)

    # execute the transformation from the vrt dataset to output compressed tif
    output_ds = gdal.Translate(output_file, ds, format="GTiff", outputType=gdal.GDT_Float32, creationOptions=["COMPRESS=DEFLATE", "TILED=YES"])
    output_ds = None
    os.remove(output_vrt)
    return




if __name__ == "__main__":
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\RSD\Alaska\Original"
    files = get_tiff_files(parent_dir, extention=".tif")

    files = [r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\RSD\Alaska\Original\ngs_dem_revillagigedo_ak_Job1105762\ngs_dem_revillagigedo_ak_J1105762_002_002.tif"]
    
    crs_from = "EPSG:6338+EPSG:5703"
    # crs_to = "EPSG:6338+NOAA:98"
    crs_to = "EPSG:6338+NOAA:55" # to match Vdatum


    files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\ALK\Original\**\*.tif", recursive=True)
    crs_from = "EPSG:6337+EPSG:5703"
    crs_to = "EPSG:6337+NOAA:98"

    steps = None

    for i, input_file in enumerate(files[:]):
        print(f"{i+1}/{len(files)}: {input_file}")
        output_file = input_file.replace("Original", "Manual")

        transform_with_vyperdatum(input_file, output_file, crs_from, crs_to, steps)

        # Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
        # transform_with_concat_pipe(input_file, output_file)
        # update_raster_wkt(output_file, pp.CRS(crs_to).to_wkt())

        # passed, cross_df = vdatum_cross_validate(s_wkt=pp.CRS(crs_from).to_wkt(),
        #                                          t_wkt=pp.CRS(crs_to).to_wkt(),
        #                                          n_sample=20,
        #                                          s_raster_metadata=raster_metadata(input_file),
        #                                          t_raster_metadata=raster_metadata(output_file),
        #                                          s_point_samples=None,
        #                                          t_point_samples=None,
        #                                          tolerance=0.3,
        #                                          raster_sampling_band=1,
        #                                          region="seak",
        #                                          pivot_h_crs="EPSG:6318",
        #                                          s_h_frame=None,
        #                                          s_v_frame=None,
        #                                          s_h_zone=None,
        #                                          t_h_frame=None,
        #                                          t_v_frame=None,
        #                                          t_h_zone=None
        #                                          )
        # cross_df.to_csv(Path(output_file).parent.absolute()/Path(f"{os.path.split(input_file)[1]}_vdatum_check.csv"), index=False)
        print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



import os
import glob
from pathlib import Path
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
import pyproj as pp
from osgeo import gdal
import re


def get_tiff_files(parent_dir: str, extention: str) -> list:
    tiff_files = []
    for (dirpath, dirnames, filenames) in os.walk(parent_dir):
        for filename in filenames:
            if filename.endswith(extention):
                tiff_files.append(os.sep.join([dirpath, filename]))
    return tiff_files


def transform_with_vyperdatum_navd88(input_file, zone):
    if zone == "16N":
        crs_from = "EPSG:6345+EPSG:5703"
        crs_to = "EPSG:6345+NOAA:93"

        crs_from = "EPSG:26916+EPSG:5703"
        crs_to = "EPSG:26916+NOAA:93"

        # crs_from = "EPSG:2251+EPSG:5703"
        # crs_to = "EPSG:2251+NOAA:93"
    elif zone == "17N":
        crs_from = "EPSG:6346+EPSG:5703"
        crs_to = "EPSG:6346+NOAA:93"
    elif zone == "18N":
        crs_from = "EPSG:6347+EPSG:5703"
        crs_to = "EPSG:6347+NOAA:93"
    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                    #  steps=steps
                     )
    output_file = input_file.replace("Original", "Manual")
    tf.transform_raster(input_file=input_file,
                        output_file=output_file,
                        overview=False,
                        pre_post_checks=True,
                        vdatum_check=False
                        )

    passed, cross_df = vdatum_cross_validate(s_wkt=pp.CRS(crs_from).to_wkt(),
                                             t_wkt=pp.CRS(crs_to).to_wkt(),
                                             n_sample=20,
                                             s_raster_metadata=raster_metadata(input_file),
                                             t_raster_metadata=raster_metadata(output_file),
                                             s_point_samples=None,
                                             t_point_samples=None,
                                             tolerance=0.3,
                                             raster_sampling_band=1,
                                             region="contiguous",
                                             pivot_h_crs="EPSG:6318",
                                             s_h_frame="NAD83_2011",
                                             s_v_frame="NAVD88",
                                             s_h_zone=None,
                                             t_h_frame="NAD83_2011",
                                             t_v_frame="LWD_IGLD85",
                                             t_h_zone=None
                                             )
    cross_df.to_csv(Path(output_file).parent.absolute()/Path(f"{os.path.split(input_file)[1]}_vdatum_check.csv"), index=False)    
    return


def transform_with_vyperdatum_igld85(input_file, zone):
    if zone == "16N":
        crs_from = "EPSG:6345+NOAA:92"
        crs_to = "EPSG:6345+NOAA:93"
        steps = [
                    {"crs_from": "EPSG:6345+NOAA:92", "crs_to": "EPSG:6318+NOAA:92", "v_shift": False},
                    {"crs_from": "EPSG:6318+NOAA:92", "crs_to": "EPSG:6318+NOAA:93", "v_shift": True},
                    {"crs_from": "EPSG:6318+NOAA:93", "crs_to": "EPSG:6345+NOAA:93", "v_shift": False}
                ]
    elif zone == "15N":
        crs_from = "EPSG:6344+NOAA:92"
        crs_to = "EPSG:6344+NOAA:93"
        steps = [
                    {"crs_from": "EPSG:6344+NOAA:92", "crs_to": "EPSG:6318+NOAA:92", "v_shift": False},
                    {"crs_from": "EPSG:6318+NOAA:92", "crs_to": "EPSG:6318+NOAA:93", "v_shift": True},
                    {"crs_from": "EPSG:6318+NOAA:93", "crs_to": "EPSG:6344+NOAA:93", "v_shift": False}
                ]
    elif zone == "17N":
        crs_from = "EPSG:6346+NOAA:92"
        crs_to = "EPSG:6346+NOAA:93"
        steps = [
                    {"crs_from": "EPSG:6346+NOAA:92", "crs_to": "EPSG:6318+NOAA:92", "v_shift": False},
                    {"crs_from": "EPSG:6318+NOAA:92", "crs_to": "EPSG:6318+NOAA:93", "v_shift": True},
                    {"crs_from": "EPSG:6318+NOAA:93", "crs_to": "EPSG:6346+NOAA:93", "v_shift": False}
                ]
    elif zone == "18N":
        crs_from = "EPSG:6347+NOAA:92"
        crs_to = "EPSG:6347+NOAA:93"
        steps = [
                    {"crs_from": "EPSG:6347+NOAA:92", "crs_to": "EPSG:6318+NOAA:92", "v_shift": False},
                    {"crs_from": "EPSG:6318+NOAA:92", "crs_to": "EPSG:6318+NOAA:93", "v_shift": True},
                    {"crs_from": "EPSG:6318+NOAA:93", "crs_to": "EPSG:6347+NOAA:93", "v_shift": False}
                ]

    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                    #  steps=steps
                     )
    output_file = input_file.replace("Original", "Manual")
    tf.transform_raster(input_file=input_file,
                        output_file=output_file,
                        overview=False,
                        pre_post_checks=True,
                        vdatum_check=False
                        )

    passed, cross_df = vdatum_cross_validate(s_wkt=pp.CRS(crs_from).to_wkt(),
                                             t_wkt=pp.CRS(crs_to).to_wkt(),
                                             n_sample=20,
                                             s_raster_metadata=raster_metadata(input_file),
                                             t_raster_metadata=raster_metadata(output_file),
                                             s_point_samples=None,
                                             t_point_samples=None,
                                             tolerance=0.3,
                                             raster_sampling_band=1,
                                             region="contiguous",
                                             pivot_h_crs="EPSG:6318",
                                             s_h_frame="NAD83_2011",
                                             s_v_frame="IGLD85",
                                             s_h_zone=None,
                                             t_h_frame="NAD83_2011",
                                             t_v_frame="LWD_IGLD85",
                                             t_h_zone=None
                                             )
    cross_df.to_csv(Path(output_file).parent.absolute()/Path(f"{os.path.split(input_file)[1]}_vdatum_check.csv"), index=False)    
    return


def transform_with_concat_pipe(input_file, output_file,
                               nwld="nwldatum_4.7.0_20240621",
                               zone=16, vdatum="MLLW"):

    output_vrt = output_file.replace(".tif", ".vrt")

    SRC_HORZ_PUSH = "+step +proj=push +v_1 +v_2"
    SRC_HORZ_POP = "+step +proj=pop +v_1 +v_2"

    NAD83_UTM_to_Geographic = lambda zone: f"+step +inv +proj=utm +zone={zone} +ellps=GRS80"
    NAD83_Geographic_to_UTM = lambda zone: f"+step +proj=utm +zone={zone} +ellps=GRS80"

    # projinfo -s EPSG:6318+EPSG:5703 -t EPSG:6319 --spatial-test intersects --hide-ballpark -o PROJ
    # (CONUS is Operation 1, AK is Operation 2)
    # AK_NAVD88_GEOID12B_to_NAD83 = "+step +proj=vgridshift +grids=us_noaa_g2012ba0.tif +multiplier=1"
    CONUS_NAVD88_GEOID18_to_NAD83 = "+step +proj=vgridshift +grids=us_noaa_g2018u0.tif +multiplier=1"


    # Note: NWLD hydroids are <datum> - [NAD83(2011) 2010.0, ITRF2020 2020.0]; to subtract, we inverse vgridshift (+step +inv)
    NAD83_2011_to_NWLD = lambda vdatum, nwld: f"+step +inv +proj=vgridshift +grids=us_noaa_nos_{vdatum}-NAD83(2011)_2010.0_({nwld}).tif +multiplier=1"

    # zone: 9, etc.
    # nwld: "nwldatum_4.7.0_20240621" (AK_ERTDM_2023), "nwldatum_4.0.0_20190729" (VDatum 4.0 SE AK only), "nwldatum_3.7.0_20170907" (AK_ERTDM_2021)
    # vdatum: "underkeel_hydroid", "MLLW", etc.

    # AK_NAD83_UTM_NAVD88_height_to_NAD83_UTM_NWLD_height = lambda zone, nwld, vdatum: f"""
    #     +proj=pipeline
    #         {NAD83_UTM_to_Geographic(zone)}
    #         {SRC_HORZ_PUSH}
    #         {AK_NAVD88_GEOID12B_to_NAD83}
    #         {NAD83_2011_to_NWLD(vdatum, nwld)}
    #         {SRC_HORZ_POP}
    #         {NAD83_Geographic_to_UTM(zone)}
    # """
    # proj_pipeline_transform = AK_NAD83_UTM_NAVD88_height_to_NAD83_UTM_NWLD_height(zone, nwld, vdatum)


    CONUS_NAD83_UTM_NAVD88_height_to_NAD83_UTM_NWLD_height = lambda zone, nwld, vdatum: f"""
        +proj=pipeline
            {NAD83_UTM_to_Geographic(zone)}
            {SRC_HORZ_PUSH}
            {CONUS_NAVD88_GEOID18_to_NAD83}
            {NAD83_2011_to_NWLD(vdatum, nwld)}
            {SRC_HORZ_POP}
            {NAD83_Geographic_to_UTM(zone)}
    """
    proj_pipeline_transform = CONUS_NAD83_UTM_NAVD88_height_to_NAD83_UTM_NWLD_height(zone, nwld, vdatum)


    # output pixel resolution = input pixel resolution
    with gdal.Open(input_file, gdal.GA_ReadOnly) as input_ds:
        geotransform = input_ds.GetGeoTransform()
        xres, yres = geotransform[1], geotransform[5]

    # create vrt transformation template dataset
    ds = gdal.Warp(output_vrt, input_file, format="vrt",
                   outputType=gdal.gdalconst.GDT_Float32,
                   warpOptions=["APPLY_VERTICAL_SHIFT=YES", "SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"],
                   errorThreshold=0, xRes=xres, yRes=yres, targetAlignedPixels=True,
                   coordinateOperation=proj_pipeline_transform)

    # remove whitespace formatting used in proj pipeline string and put it in GeoTIFF metadata
    proj_pipeline_transform = re.sub(r'\s{2,}', ' ', proj_pipeline_transform).strip()
    ds.SetMetadataItem('TIFFTAG_IMAGEDESCRIPTION', proj_pipeline_transform)

    # execute the transformation from the vrt dataset to output compressed tif
    output_ds = gdal.Translate(output_file, ds, format="GTiff",
                               outputType=gdal.GDT_Float32,
                               creationOptions=["COMPRESS=DEFLATE", "TILED=YES"])
    output_ds = None
    os.remove(output_vrt)
    return


import sys
input_file = r"C:\Users\mohammad.ashkezari\Desktop\Fuse_Test\raster\Original\ny\ny2018_ont_dem_J1145472_004_003.tif"
transform_with_vyperdatum_navd88(input_file, "18N")
sys.exit()

if __name__ == "__main__":
    # files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBD\Original\NAVD88\16N\**\*.tif", recursive=True)
    # files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBD\Original\NAVD88\**\*.tif", recursive=True)
    # files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBD\Original\IGLD85\**\*.tif", recursive=True)

    files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBD\Original\NAVD88\16N\usace2009_isle_roy_dem_Job1201041\*.tif", recursive=True)
    # files = [r"C:\Users\mohammad.ashkezari\Desktop\Fuse_Test\raster\Original\ny\ny2018_ont_dem_J1145472_005_003.tif"]

    for i, input_file in enumerate(files[:]):
        try:
            print(f"{i+1}/{len(files)}: {input_file}")
            if input_file.find(r"\16N") != -1:
                zone = "16N"
            elif input_file.find(r"\15N") != -1:
                zone = "15N"
            elif input_file.find(r"\17N") != -1:
                zone = "17N"
            elif input_file.find(r"\18N") != -1:
                zone = "18N"
            transform_with_vyperdatum_navd88(input_file, zone)
            # transform_with_vyperdatum_igld85(input_file, zone)

            print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')
        except Exception as e:
            print(f"Error transforming {input_file}:\n{e}")


import requests
import logging
import time
import random
import numpy as np
import pandas as pd
from colorama import Fore, Style
from osgeo import gdal
from typing import Optional, Union
import pyproj as pp
from tqdm import tqdm
import difflib
from vyperdatum.enums import VDATUM


logger = logging.getLogger("root_logger")
gdal.UseExceptions()


def vdatum_transform_point(s_x, s_y, s_z, region,
                           s_h_frame, s_v_frame, s_h_zone,
                           t_h_frame, t_v_frame, t_h_zone,
                           s_v_goid="geoid18",
                           t_v_goid="geoid18",
                           ):
    """
    Call point transformation GET endpoint (/convert) of the Vdatum REST API.
    API docs: https://vdatum.noaa.gov/docs/services.html

    Parameters
    ----------
    s_x: float
        Source point longitude
    s_y: float
        Source point latitude
    s_z: float
        Source point height
    region: str
        Vdatum region name (e.g. ak, as, contiguous, gcnmi, prvi)
        https://vdatum.noaa.gov/docs/services.html#step140
    s_h_frame: str
        Source horizontal reference frame (e.g. NAD83_2011)
        https://vdatum.noaa.gov/docs/services.html#step150
    s_v_frame: str
        source vertical reference frame (e.g. NAVD88, PRVD02, MLLW)
        https://vdatum.noaa.gov/docs/services.html#step160
    t_h_frame: str
        Target horizontal reference frame (e.g. NAD83_2011)
        https://vdatum.noaa.gov/docs/services.html#step150
    t_v_frame: str
        Target vertical reference frame (e.g. NAVD88, PRVD02, MLLW)
        https://vdatum.noaa.gov/docs/services.html#step160
    s_v_goid: str
        Source vertical GEOID model (e.g. geoid18 (default), geoid12b,
        geoid12a, geoid09, geoid06, geoid03, geoid99, geoid96, egm2008,
        egm1996, egm1984, xgeoid16b, xgeoid17b, xgeoid18b, xgeoid19b,
        xgeoid20b)
    t_v_goid: str
        Target vertical GEOID model (e.g. geoid18 (default), geoid12b,
        geoid12a, geoid09, geoid06, geoid03, geoid99, geoid96, egm2008,
        egm1996, egm1984, xgeoid16b, xgeoid17b, xgeoid18b, xgeoid19b,
        xgeoid20b)

    Returns
    -------
    Optional[tuple]
        The transformed coordinates (lon, lat, height)
    Optional[dict]
        The complete vdatum response object
    """
    url = "https://vdatum.noaa.gov/vdatumweb/api/convert"
    params = dict(s_x=s_x, s_y=s_y, s_z=s_z, region=region,
                  s_h_frame=s_h_frame, s_v_frame=s_v_frame,
                  s_h_zone=s_h_zone if s_h_zone else "",
                  t_h_frame=t_h_frame, t_v_frame=t_v_frame,
                  t_h_zone=t_h_zone if t_h_zone else "",
                  s_v_goid=s_v_goid, t_v_goid=t_v_goid)
    # logger.debug(f"Parameter used to call Vdatum API:\n{params}\n")
    try:
        resp = requests.get(url, params=params, timeout=200).json()
        time.sleep(0.5)
    except Exception as e:
        logger.exception(f"VDatum API exception:\n{e}")
        return None, None
    return (float(resp.get("t_x", np.nan)),
            float(resp.get("t_y", np.nan)),
            float(resp.get("t_z", np.nan))), resp


def api_region_alias(vdatum_regions: list[str]):
    """
    Vdatum REST api uses different region names compared to those
    listed in the vdatum grids directory. This function expects to
    receive region name according to the vdatum grids directory and
    returns region name consistent with the vdatum REST api. Default
    output is `contiguous`.

    Parameters
    ------------
    region: list[str]
         Region name according to the vdatum grids directory.

    Returns
    ------------
    str
    """
    if not vdatum_regions:
        return "contiguous"
    api_region = set()
    for region in vdatum_regions:
        if region.startswith("AK"):
            api_region.add("ak")
        elif region.startswith("MD"):
            api_region.add("chesapeak_delaware")
        elif region.startswith("TX"):
            api_region.add("wgom")
        elif region.startswith("WA") or region.startswith("OR") or region.startswith("WC"):
            api_region.add("westcoast")
        elif region.startswith("PRVI"):
            api_region.add("prvi")

    if len(api_region) != 1:
        api_region = "contiguous"
    else:    
        api_region = api_region.pop()
    # Note: not all regions are covered here. For example, I do'nt know what is the Hawaii
    # region name is in vdatum grid directory (or as, sgi, spi, sli, gcnmi, wgom, ...).
    # looks like `westcoast` and `chesapeak_delaware` are not considered as part of `contiguous`?.
    # https://vdatum.noaa.gov/docs/services.html#step140
    return api_region


def wkt_to_utm(wkt: str) -> Optional[str]:
    """
    Receives WKT and returns the UTM zone, if applicable.

    Parameters
    ------------
    wkt: str
         WKT string

    Returns
    ------------
    Optional[str]
         UTM zone
    """
    zone = None
    input_crs = pp.CRS(wkt)
    if not input_crs.is_projected:
        return zone
    input_crs = input_crs.name.split("+")[0].strip()
    if len(input_crs.split("/ UTM zone ")) == 2:
        zone = input_crs.split("/ UTM zone ")[1].strip()
    return zone


def wkt_to_crs(wkt: str) -> tuple[Optional[str], Optional[str]]:
    """
    Receives WKT and returns the horizontal/vertical CRS names.

    Parameters
    ------------
    wkt: str
         WKT string

    Returns
    ------------
    tuple[Optional[str], Optional[str]]
         Horizontal and vertical CRS names, if identified.
    """
    input_crs = pp.CRS(wkt)
    if input_crs.is_compound:
        horizontal_crs = pp.CRS(input_crs.sub_crs_list[0])
        vertical_crs = pp.CRS(input_crs.sub_crs_list[1])
    else:
        horizontal_crs = input_crs
        vertical_crs = None
    return horizontal_crs.name, vertical_crs.name if vertical_crs else None


def api_crs_aliases(wkt: str) -> tuple[str, str]:
    """
    Vdatum REST api uses different CRS names compared to those
    of the PROJ/pyproj. This function expects to receive a CRS WKT
    and returns horizontal and vertical CRS names that are consumable
    by the vdatum REST api.


    Parameters
    ------------
    wkt: str
         WKT string.

    Returns
    ------------
    tuple[str, str]
        Horizontal and vertical CRS names that are consumable by the vdatum REST api.
    """
    h_crs, v_crs = wkt_to_crs(wkt)
    h_crs = h_crs.split("/")[0].strip()
    h_matches = difflib.get_close_matches(h_crs, VDATUM.H_FRAMES.value)
    if len(h_matches) == 0:
        logger.error(f"No Vdatum horizontal CRS name matched with '{h_crs}'")
    h_crs = h_matches[0]

    v_crs_black = ["HRD"]
    if v_crs and v_crs.split()[0].upper().strip() in v_crs_black:
        logger.error(f"The raster's vertical CRS is '{v_crs}' which is not covered by Vdatum.")

    if not v_crs:
        v_crs = h_crs
    else:
        v_matches = difflib.get_close_matches(v_crs, VDATUM.V_FRAMES.value)
        if len(v_matches) == 0:
            v_matches = difflib.get_close_matches(v_crs.split()[0].strip(), VDATUM.V_FRAMES.value)
        if len(v_matches) == 0:
            v_matches = difflib.get_close_matches(v_crs.split()[0].strip(), VDATUM.V_FRAMES.value)
            logger.error(f"No Vdatum vertical CRS name matched with '{v_crs}'")
        v_crs = v_matches[0]
    return h_crs, v_crs


def index_to_xy(i: int, j: int, geot: tuple):
    """
    Take indices of a raster's band array and return x, y.

    Parameters
    ----------
    i: int
        Raster's band array first index.
    j: int
        Raster's band array second index.
    geot: tuple
        Gdal GeoTransform tuple object.

    Returns
    ----------
    float, float
        easting, northing
    """
    res_x, res_y = geot[1], geot[5]
    x_min, y_max = geot[0], geot[3]
    x = j * res_x + x_min + (res_x / 2.0)
    y = i * res_y + y_max + (res_y / 2.0)
    return x, y


def sample_raster(source_meta: dict,
                  target_meta: dict,
                  n_sample: int,
                  sampling_band: int,
                  pivot_h_crs: Optional[str]
                  ) -> tuple[list, list]:
    """
    Randomly draw `n_samples` points (not NoDataValue) from the `sampling_band`
    of the source and target rasters.

    Parameters
    ----------
    source_meta: dict
        Source raster metadata generated by `raster_metadata` function.
    target_meta: dict
        Target raster metadata generated by `raster_metadata` function.
    n_sample: int
        The number of sample points (coordinates).
    sampling_band: dict
        The index of the source band to be sampled (default 1).
    pivot_h_crs: Optional[str]
        When not None, both source and target samples are transformed
        into a common horizontal CRS, otherwise ignored (default None).
        Example: pivot_h_crs = 'EPSG:6318'

    Raises
    -------
    ValueError:
        If there are less not-null values than `n_sample` in the target raster band.

    Returns
    ----------
    tuple[list, list]
        Returns `n_samples` points from the rasters in form of two lists of [x, y, sampled_value].
    """
    ds_source = gdal.Open(source_meta["path"])
    ds_target = gdal.Open(target_meta["path"])
    bar_source = ds_source.GetRasterBand(sampling_band).ReadAsArray()
    bar_target = ds_target.GetRasterBand(sampling_band).ReadAsArray()
    nodata_value = ds_target.GetRasterBand(sampling_band).GetNoDataValue()
    if np.count_nonzero(np.where(bar_target != nodata_value)) < n_sample:
        raise ValueError(f"Cannot sample {n_sample} from the target raster. Too many NoDataValue.")
    source_samples, target_samples = [], []
    c = 0
    while len(target_samples) < n_sample:
        c += 1
        if c > 1e6:
            raise ValueError(f"Failed to find {n_sample} sample points from the target raster.")
        i, j = np.random.choice(bar_target.shape[0]), np.random.choice(bar_target.shape[1])
        if bar_target[i, j] == nodata_value:
            continue
        x, y = index_to_xy(i, j, ds_target.GetGeoTransform())
        target_samples.append([x, y, bar_target[i, j]])

        source_geot = ds_source.GetGeoTransform()
        source_i = int((y - source_geot[3]) / source_geot[5])
        source_j = int((x - source_geot[0]) / source_geot[1])

        source_x, source_y = index_to_xy(source_i, source_j, ds_source.GetGeoTransform())
        source_samples.append([source_x, source_y, bar_source[source_i, source_j]])
    ds_source, ds_target = None, None
    if pivot_h_crs:
        source_crs_h, _ = wkt_to_crs(source_meta["wkt"])
        target_crs_h, _ = wkt_to_crs(target_meta["wkt"])
        source_transformer = pp.Transformer.from_crs(pp.CRS(source_crs_h),
                                                     pp.CRS(pivot_h_crs),
                                                     always_xy=True
                                                     )
        target_transformer = pp.Transformer.from_crs(pp.CRS(target_crs_h),
                                                     pp.CRS(pivot_h_crs),
                                                     always_xy=True
                                                     )
        for i, p in enumerate(source_samples):
            lon, lat = source_transformer.transform(p[0], p[1])
            p[0], p[1] = lon, lat
            lon, lat = target_transformer.transform(target_samples[i][0], target_samples[i][1])
            target_samples[i][0], target_samples[i][1] = lon, lat
    return source_samples, target_samples


def vdatum_cross_validate(s_wkt: str,
                          t_wkt: str,
                          n_sample: int,
                          s_raster_metadata: Optional[dict],
                          t_raster_metadata: Optional[dict],
                          s_point_samples: Optional[Union[list, np.ndarray]],
                          t_point_samples: Optional[Union[list, np.ndarray]],
                          tolerance: float = 0.3,
                          raster_sampling_band: int = 1,
                          region: Optional[str] = None,
                          pivot_h_crs: Optional[str] = None,
                          s_h_frame: Optional[str] = None,
                          s_v_frame: Optional[str] = None,
                          s_h_zone: Optional[str] = None,
                          t_h_frame: Optional[str] = None,
                          t_v_frame: Optional[str] = None,
                          t_h_zone: Optional[str] = None
                          ):
    """
    Randomly sample the source raster points and transform them to the
    target CRS using the vdatum API. Verify if the transformed values are
    consistent with the target raster values.

    Parameters
    ----------
    s_wkt: str
        WKT associated with the source data CRS.
    t_wkt: str
        WKT associated with the target data CRS.
    n_sample: int
        The number of sample points (coordinates).
    s_raster_metadata: Optional[dict]
        Source file metadata. This only applies to raster files; otherwise, None.
    t_raster_metadata: Optional[dict]
        Target file metadata. This only applies to raster files; otherwise, None.
    s_point_samples: Optional[Union[list, numpy.ndarray]]
        Lists of source sample points in form of [x, y, sampled_value].
        This parameter only applies to point transformations(not raster); otherwise, None.
    t_point_samples: Optional[Union[list, numpy.ndarray]]
        Lists of source sample points in form of [x, y, sampled_value].
        This parameter only applies to point transformations(not raster); otherwise, None.
    tolerance: float
        Maximum elevation deviation [m] between transformations conducted by Vyperdatum
        compared with Vdatum.
    raster_sampling_band: dict
        The index of the source raster file band to be sampled (default 1). Only applies to
        raster file inputs.
    region: Optional[str]
        Vdatum region name (e.g. ak, as, contiguous, gcnmi, prvi)
        https://vdatum.noaa.gov/docs/services.html#step140
    pivot_h_crs: Optional[str]
        When not None, both source and target samples are transformed
        into a common horizontal CRS, otherwise ignored (default None).
        Example: pivot_h_crs = 'EPSG:6318'
    s_h_frame: Optional[str]
        Source horizontal reference frame (e.g. NAD83_2011)
        https://vdatum.noaa.gov/docs/services.html#step150
    s_v_frame: Optional[str]
        source vertical reference frame (e.g. NAVD88, PRVD02, MLLW)
        https://vdatum.noaa.gov/docs/services.html#step160
    s_h_zone: Optional[str]
        source CRS zone, if projected.
        The details of this parameter is not documented in vdatum services page.
    t_h_frame: Optional[str]
        Target horizontal reference frame (e.g. NAD83_2011)
        https://vdatum.noaa.gov/docs/services.html#step150
    t_v_frame: Optional[str]
        Target vertical reference frame (e.g. NAVD88, PRVD02, MLLW)
        https://vdatum.noaa.gov/docs/services.html#step160
    t_h_zone: Optional[str]
        Target CRS zone, if projected.
        The details of this parameter is not documented in vdatum services page.

    Returns
    ----------
    bool
        True if all checks pass, otherwise False.
    pandas.DataFrame
        Dataframe containing the sampled input, transferred, and vdatum points.
    """
    # source_meta = raster_metadata(s_file)
    # target_meta = raster_metadata(t_file)
    # s_wkt = source_meta["wkt"]
    # t_wkt = target_meta["wkt"]

    passed = True
    if not region:
        overlapping_regions = t_raster_metadata.get("overlapping_regions") if s_raster_metadata else None
        region = api_region_alias(overlapping_regions)
        # if len(s_raster_metadata["overlapping_regions"]) != 1:
        #     logger.warning("The input is not overlapping with a single region."
        #                    f" The overlapping regions: ({s_raster_metadata['overlapping_regions']})."
        #                    " The Vdatum API region will be set to 'contiguous'.")
            

    source_crs_h, source_crs_v = wkt_to_crs(s_wkt)
    source_zone_h = wkt_to_utm(s_wkt)
    target_crs_h, target_crs_v = wkt_to_crs(t_wkt)
    target_zone_h = wkt_to_utm(t_wkt)
    s_h_zone = s_h_zone if s_h_zone else source_zone_h
    t_h_zone = t_h_zone if t_h_zone else target_zone_h
    if pivot_h_crs:
        s_h_zone, t_h_zone = None, None

    api_src_crs_h, api_src_crs_v = api_crs_aliases(s_wkt)
    source_crs_h = s_h_frame if s_h_frame else api_src_crs_h
    source_crs_v = s_v_frame if s_v_frame else api_src_crs_v
    api_tgt_crs_h, api_tgt_crs_v = api_crs_aliases(t_wkt)
    target_crs_h = t_h_frame if t_h_frame else api_tgt_crs_h
    target_crs_v = t_v_frame if t_v_frame else api_tgt_crs_v

    if s_raster_metadata and t_raster_metadata:  # in the case of raster file inputs
        logger.info(f"Taking {n_sample} random sample{'s' if n_sample > 1 else ''} "
                    "from the source and transformed rasters ...")
        source_samples, target_samples = sample_raster(s_raster_metadata,
                                                       t_raster_metadata,
                                                       n_sample,
                                                       raster_sampling_band,
                                                       pivot_h_crs=pivot_h_crs
                                                       )
    else:
        random_idx = random.sample(range(len(s_point_samples)), n_sample)
        source_samples = [list(s_point_samples[i]) for i in random_idx]
        target_samples = [list(t_point_samples[i]) for i in random_idx]

    vdatum_points, vdatum_resp = [], []
    cross_df = pd.DataFrame({})
    logger.info("Calling Vdatum API ...")
    for p in tqdm(source_samples):
        points, resp = vdatum_transform_point(s_x=p[0], s_y=p[1], s_z=p[2],
                                              region=region,
                                              s_h_frame=s_h_frame or source_crs_h,
                                              s_v_frame=s_v_frame or source_crs_v,
                                              s_h_zone=s_h_zone,
                                              t_h_frame=t_h_frame or target_crs_h,
                                              t_v_frame=t_v_frame or target_crs_v,
                                              t_h_zone=t_h_zone)
        if points:
            vdatum_points.append(points)
            vdatum_resp.append(resp)
        else:
            vdatum_points.append((None, None, None))
            vdatum_resp.append({})
    cross_df = pd.DataFrame({"src_x": [p[0] for p in source_samples],
                             "src_y": [p[1] for p in source_samples],
                             "src_val": [p[2] for p in source_samples],
                             "tgt_x": [p[0] for p in target_samples],
                             "tgt_y": [p[1] for p in target_samples],
                             "tgt_val": [p[2] for p in target_samples],
                             "vdatum_x": [p[0] for p in vdatum_points],
                             "vdatum_y": [p[1] for p in vdatum_points],
                             "vdatum_val": [p[2] for p in vdatum_points],
                             "vdatum_resp": [str(r) for r in vdatum_resp],
                             })
    try:
        # remove vdatum outlier/erroneous responses
        cross_df = cross_df.query("abs(vdatum_val)<10000")
        if len(cross_df) == 0:
            passed = False
        cross_df = cross_df.assign(deviation=cross_df["tgt_val"] - cross_df["vdatum_val"])
        deviation = cross_df["deviation"].abs().mean()
        std = cross_df["deviation"].abs().std()
        if deviation > tolerance:
            logger.warning(f"{Fore.RED}The deviation between the transferred values produced by "
                           f"Vyperdatum and Vdatum is {deviation:.2f} {chr(0x00b1)} {std:.2f} [m]"
                           f" exceeding the threshold {tolerance} [m].")
            passed = False
        elif deviation != deviation:
            logger.warning(f"{Fore.RED}The deviation between the transferred values produced by "
                           f"Vyperdatum and Vdatum is ambiguous (nan).")
            passed = False
        else:
            logger.warning(f"{Fore.GREEN}The deviation between the transferred values produced by "
                           f"Vyperdatum and Vdatum is {deviation:.2f} {chr(0x00b1)} {std:.2f} [m],"
                           f" below the threshold {tolerance} [m].")
        print(Style.RESET_ALL)
    except Exception as e:
        logger.exception(f"Exception in Vdatum deviation calculation:\n{e}")
        passed, cross_df = False, pd.DataFrame({})
    return passed, cross_df



import os
import logging
from osgeo import gdal
import geopandas as gpd
from shapely.geometry import Point
import numpy as np
import pyproj as pp
from vyperdatum.utils.crs_utils import auth_code
from vyperdatum.drivers.base import Driver

logger = logging.getLogger("root_logger")
gdal.UseExceptions()


class GeoParquet(Driver):
    def __init__(self, input_file: str, invalid_error: bool = True) -> None:
        """
        Parameters
        -----------
        input_file: str
            Path to the input geoparquet file.
        invalid_error: bool, default True
            If True, throws an error when the input file has an unexpected format.

        Raises
        -------
        FileNotFoundError:
            If the input file is not found.
        ValueError
            If the passed input file is not recognized as geoparquet file.

        Returns
        -----------
        None
        """
        super().__init__()
        self.input_file = input_file
        if not os.path.isfile(self.input_file):
            raise FileNotFoundError(f"The input file not found at {input_file}.")
        self.is_gparq = self._is_valid()
        if invalid_error and not self.is_gparq:
            msg = (f"The following file is not a valid geoparquet point file: {self.input_file}")
            logger.exception(msg)
            raise TypeError(msg)
        return

    def _is_valid(self) -> bool:
        """
        Check if the input file is a geoparquet and only contains points.

        Returns
        -----------
        bool
            True if a valid geoparquet with point data; otherwise False.
        """
        try:
            gdf = gpd.read_parquet(self.input_file)
            geotypes = gdf.geom_type.unique()
            if len(geotypes) != 1 or geotypes[0].lower() != "Point".lower():
                return False
            valid = True
        except:
            valid = False
        return valid

    def get_points(self) -> bool:
        """
        Extract the coordinate points from geoparquet file.
        The input file should be in the geoparquet format and should only contain points.

        Returns
        -----------
        bool
            True if the coordinates are found successfully; otherwise False.
        """
        try:
            if not self._is_valid():
                return False
            gdf = gpd.read_parquet(self.input_file)
            logger.info("Reading geoparquet file into numpy arrays ...")
            coords = np.array([geom.coords[0] for geom in gdf.geometry])
            self.x, self.y, self.z = coords[:, 0], coords[:, 1], coords[:, 2]
            valid = True
        except:
            valid = False
        return valid

    def wkt(self) -> str:
        """
        Return the WKT string.

        Returns
        -----------
        str
            WKT associated with file's CRS.
        """
        gdf = gpd.read_parquet(self.input_file)
        if gdf.crs is None:
            raise ValueError("The input file does not have a CRS.")
        w = gdf.crs.to_wkt()
        return w

    def transform(self, transformer_instance, output_file: str, pre_post_checks: bool, vdatum_check: bool) -> bool:
        """
        Apply point transformation on the geoparquet point data according to the `transformer_instance`.

        Parameters
        -----------
        transformer_instance: vyperdatum.transformer.Transform
            Instance of the transformer class.
        output_file: str
            Path to the output geoparquet file.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        if not hasattr(self, "x"):
            self.get_points()
        gdf = gpd.read_parquet(self.input_file)
        if pre_post_checks:
            source_crs = transformer_instance.crs_from
            if not isinstance(transformer_instance.crs_from, pp.CRS):
                source_crs = pp.CRS(transformer_instance.crs_from)
            source_auth = auth_code(source_crs)
            file_auth = auth_code(pp.CRS(gdf.crs.to_wkt()))
            if source_auth != file_auth:
                logger.warning("The authority name/code registered in the "
                               f"input file is {file_auth}, but expected {source_auth}"
                               )
        success, xt, yt, zt = transformer_instance.transform_points(self.x,
                                                                    self.y,
                                                                    self.z,
                                                                    vdatum_check=vdatum_check
                                                                    )
        new_geometry = [Point(xi, yi, zi) for xi, yi, zi in zip(xt, yt, zt)]
        new_gdf = gpd.GeoDataFrame(gdf[["Uncertainty", "Classification"]].copy(),
                                   geometry=new_geometry,
                                   crs=pp.CRS(transformer_instance.crs_to)
                                   )
        new_gdf.to_parquet(output_file, compression="gzip", index=False)

        if pre_post_checks:
            new_gdf = gpd.read_parquet(output_file)
            target_crs = transformer_instance.crs_to
            if not isinstance(transformer_instance.crs_to, pp.CRS):
                target_crs = pp.CRS(transformer_instance.crs_to)
            target_auth = auth_code(target_crs)
            transformed_file_auth = auth_code(pp.CRS(new_gdf.crs.to_wkt()))
            if target_auth != transformed_file_auth:
                logger.warning("The expected authority name/code of the "
                               f"transformed geoparquet is {target_auth}, but received {transformed_file_auth}"
                               )        
        return success

    @property
    def is_valid(self):
        return self.is_gparq
